//	Ball.h - The ball class.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef BALL_H_
#define BALL_H_

#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "Collidable.h"

class Drawer;

///	The ball class.
class Ball : public Collidable
{
  public:
	///	Constructor.
	Ball(const TwoFloats& pos);
	///	Destructor.
	~Ball();

	///	Moves the ball.
	void update();
	///	Ensures the ball stays within the arena.
	void arenaWallCheck(const TwoFloats& screenSize);
	///	Draws the ball.
	void draw(shared_ptr<Drawer>& d,
			  float timestepInterp,
			  const ThreeFloats& leftCol,
			  const ThreeFloats& rightCol);

	///	Resets the ball.
	void reset(const TwoFloats& pos);
	///	Puts the ball into slow motion.
	void setSlow();

	///	Kicks the ball.
	void kick(const TwoFloats& dir);

	///	Updates the ball's speed.
	/*!
		\param val 0->1
	 */
	void updateSpeed(float val);

	///	Returns true if this entity wants to know when it collides with the named entity type.
	bool interestedInCollisionsWith(const std::wstring& entity) const;
	///	Called when this entity collides with another.
	bool collide(const shared_ptr<Collidable>& other, GamePlayState& state);

	///	Returns the entity's position.
	const TwoFloats& getPosition() const {return position;};
	///	Returns the entity's speed.
	float getSpeed() const {return direction.getMagnitude();};
	///	Returns the entity's radius.
	float getRadius() const {return 12.0f;};
	///	Returns true if other entities can pass through this one.
	bool getIncorporeal() const {return false;};
  private:
	///	The length of the ball's tail.
	enum
	{
		TailSize = 32
	};

	///	The ball's position.
	TwoFloats position;
	///	The ball's direction vector.
	TwoFloats direction;
	///	The ball's speed.
	float speed;
	///	The ball's size.
	TwoFloats size;
	///	The ball's size tweener.
	LogTweener<TwoFloats> sizeTweener;

	///	The ball's tail.
	std::vector<TwoFloats> tailPos;
	///	The ball's tail.
	std::vector<float> tailSize;
	///	The ball's tail.
	std::vector<ThreeFloats> tailCol;
	///	The ball's tail.
	std::vector<float> tailAlpha;

	///	The flame images.
	std::wstring flameImages[6];
	///	The y-axis offsets of the flame images.
	float flameOffsets[6];
	///	Which flame image we're on.
	int flameIndex;
	///	Used to update flameIndex.
	int flameCount;
	///	Used to interpolate the flame colours.
	ThreeFloats newCol;
	///	Used to interpolate the flame colours.
	ThreeFloats oldCol;
	///	Used to interpolate the flame colours.
	float flameInterp;
	///	The flame particles.
	ParticleGenerator particles;
};

#endif
