//	Player.h - Class representing one of the players.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PLAYER_H_
#define PLAYER_H_

#include "ParticleGenerator.h"
#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "Collidable.h"

#include <vector>
#include <deque>

class Drawer;

///	Class representing one of the players.
class Player : public Collidable
{
  public:
	///	Constructor.
	Player(const TwoFloats& pos, float ang, const ThreeFloats& col);
	///	Destructor.
	~Player();

	///	Moves & animates the player.
	void update();
	///	Draws the player's tail.
	void drawTail(shared_ptr<Drawer>& d, float timestepInterp);
	///	Draws the player.
	void draw(shared_ptr<Drawer>& d, float timestepInterp);

	///	Resets the player to its starting position.
	void reset(const TwoFloats& pos, float ang);

	///	Increments/decrements the length of the player's tail.
	void incTail(int numNewSegments);
	///	Sets the player's power level.
	void setPowerLevel(float val);
	///	Toggles the player's tail flash.
	void setTailFlash(bool val);

	///	Called when a joysick button/key down event is received for this player.
	void keyDown(const std::wstring& action);
	///	Called when a joysick button/key up event is received for this player.
	void keyUp(const std::wstring& action);
	///	Called when a joystick move event is received for this player.
	void move(const std::wstring& action, float value);

	///	Returns the player's tail.
	const std::vector<TwoFloats>& getTail() const {return tail;};

	///	Returns true if this entity wants to know when it collides with the named entity type.
	bool interestedInCollisionsWith(const std::wstring& entity) const;
	///	Called when this entity collides with another.
	bool collide(const shared_ptr<Collidable>& other, GamePlayState& state);

	///	Returns the entity's position.
	const TwoFloats& getPosition() const {return position;};
	///	Returns the entity's speed.
	float getSpeed() const {return direction.getMagnitude() * speed;};
	///	Returns the entity's radius.
	float getRadius() const;
	///	Returns true if other entities can pass through this one.
	bool getIncorporeal() const {return false;};
  private:
	///	Some constants.
	enum
	{
		StartingTailSize = 96
	};

	///	The player's position.
	TwoFloats position;
	///	The player's colour.
	ThreeFloats colour;
	///	The player's tail.
	std::vector<TwoFloats> tail;
	///	The angle the player's facing.
	float angle;
	///	The player's current direction.
	TwoFloats direction;
	///	The player's desired direction.
	TwoFloats desiredDirection;
	///	The player's speed.
	float speed;
	///	The player's power level.
	float powerLevel;

	///	Used to move the player's arms when they're running.
	float armPos;

	///	Test.
	TwoFloats collidePosition;

	///	Temporary deque used to update tail when its size changes.
	std::deque<TwoFloats> tempTail;

	///	For drawing the player's tail.
	std::vector<float> tailSize;
	///	For drawing the player's tail.
	std::vector<ThreeFloats> tailCol;
	///	For drawing the player's tail.
	std::vector<float> tailAlpha;

	///	True if the player's tail should flash (because it won't collide).
	bool tailFlash;
	///	Used to flash the player's tail.
	int tailFlashCount;
	///	The flame images.
	std::wstring flameImages[6];
	///	The y-axis offsets of the flame images.
	float flameOffsets[6];
	///	Which flame image we're on.
	int flameIndex;
	///	Used to update flameIndex.
	int flameCount;
	///	Used to interpolate the flame colours.
	ThreeFloats newCol;
	///	Used to interpolate the flame colours.
	ThreeFloats oldCol;
	///	Used to interpolate the flame colours.
	float flameInterp;
	///	The flame particles.
	ParticleGenerator particles;

	///	For moving the player with the keyboard.
	bool upIsDown;
	///	For moving the player with the keyboard.
	bool downIsDown;
	///	For moving the player with the keyboard.
	bool leftIsDown;
	///	For moving the player with the keyboard.
	bool rightIsDown;
};

#endif
