//	SnakeballSounder.cpp - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "sound/SoundFileLoader.h"
#include "SnakeballSounder.h"
#include "DebugHeaders.h"
#include "MiscHelper.h"

#include <cassert>

using std::wstring;
using std::vector;
using std::map;

//------------------------------------------------------------------------------
SnakeballSounder::SnakeballSounder():
Thread("SnakeballSounder"),
inputPos(0),
music(-1),
timeToStop(false),
numChannels(2),
redInputChannel(0),
blueInputChannel(0),
redInput(0.0f),
blueInput(0.0f),
shouting(false)
{
	int i;

	for(i=0;i<InputBufferSize;++i)
	{
		inputBuffer[0][i] = 0.0f;
		inputBuffer[1][i] = 0.0f;
	}
	for(i=0;i<NumSounds;++i)
		envelopes[i].inc = 0.0f;

	delay.setDelay(0.25f);
	delay.setFeedback(0.66f);
	delay.setMix(0.125f);

	startThread(this);
}

//------------------------------------------------------------------------------
SnakeballSounder::~SnakeballSounder()
{
	map<wstring, PreLoadedBuffer>::iterator it;

	timeToStop = true;
	Thread::wait(300);

	for(it=preLoadedBuffers.begin();
		it!=preLoadedBuffers.end();
		++it)
	{
		delete [] it->second.buffer;
	}
}

//------------------------------------------------------------------------------
void SnakeballSounder::getAudio(float *input, float *output, int numSamples)
{
	int i, j;
	float tempf;
	int readPos;
	float leftEnv, rightEnv;

	//Update the red & blue input levels, write input audio to inputBuffer.
	for(i=0,j=inputPos;i<numSamples;++i,++j)
	{
		tempf = fabsf(input[(i*numChannels)+redInputChannel]);
		if(tempf >= redInput)
			redInput = tempf;
		else
		{
			redInput -= 0.00001f;
			if(redInput < 0.0f)
				redInput = 0.0f;
		}
	
		tempf = fabsf(input[(i*numChannels)+blueInputChannel]);
		if(tempf >= blueInput)
			blueInput = tempf;
		else
		{
			blueInput -= 0.00001f;
			if(blueInput < 0.0f)
				blueInput = 0.0f;
		}

		j %= InputBufferSize;

		inputBuffer[0][j] = input[(i*numChannels)+redInputChannel];
		inputBuffer[1][j] = input[(i*numChannels)+blueInputChannel];
	}

	//Clear the buffer, just in case.
	for(i=0;i<(numSamples*2);++i)
		output[i] = 0.0f;
	
	//Play any sounds.
	for(i=0;i<NumSounds;++i)
	{
		if(sounds[i].getActive())
			sounds[i].getAudio(output, numSamples);
	}

	//Play/update any input envelopes.
	if(shouting)
		readPos = inputPos;
	else
		readPos = (inputPos+numSamples+1)%InputBufferSize;

	for(i=0;i<numSamples;++i)
	{
		leftEnv = 0.0f;
		rightEnv = 0.0f;

		for(j=0;j<NumSounds;++j)
		{
			if(envelopes[j].inc != 0.0f)
			{
				envelopes[j].index += envelopes[j].inc;
				if(envelopes[j].index >= 1.0f)
				{
					envelopes[j].index = 1.0f;
					envelopes[j].inc = -0.0002f;
				}
				else if(envelopes[j].index <= 0.0f)
				{
					envelopes[j].index = 0.0f;
					envelopes[j].inc = 0.0f;
				}

				tempf = fakeLog(envelopes[j].index);
				leftEnv += tempf * envelopes[j].left * envelopes[j].level;
				rightEnv += tempf * envelopes[j].right * envelopes[j].level;
			}
		}

		if(leftEnv > 1.0f)
			leftEnv = 1.0f;
		if(rightEnv > 1.0f)
			rightEnv = 1.0f;

		output[(i*2)] += inputBuffer[0][readPos] * leftEnv;
		output[(i*2)+1] += inputBuffer[1][readPos] * rightEnv;

		++inputPos;
		inputPos %= InputBufferSize;
		++readPos;
		readPos %= InputBufferSize;
	}

	//Apply the delay.
	delay.getAudio(output, numSamples);
}

//------------------------------------------------------------------------------
void SnakeballSounder::loadSoundFileResource(SoundFileLoaderPtr soundFileLoader)
{
	int i, j;
	const int64_t fileLength = soundFileLoader->getFileLength();
	const int64_t shortFileMaxSize = (2 * 44100);

	soundFiles.insert(make_pair(soundFileLoader->getName(), soundFileLoader));

	if(fileLength < shortFileMaxSize)
	{
		int64_t bufPos;
		PreLoadedBuffer buffer;

		buffer.buffer = new float[(uint32_t)fileLength * 2];
		buffer.size = (int)fileLength * 2;

		bufPos = 0;
		while(bufPos < fileLength)
		{
			bufPos = soundFileLoader->loadChunk(buffer.buffer,
												bufPos,
												buffer.size,
												0);
		}

		//Convert mono files to stereo.
		if(soundFileLoader->getNumChannels() < 2)
		{
			for(j=(buffer.size-2);j>=0;j-=2)
			{
				buffer.buffer[j] = buffer.buffer[j/2];
				buffer.buffer[j+1] = buffer.buffer[j];
			}
		}

		preLoadedBuffers.insert(make_pair(soundFileLoader->getName(), buffer));
	}
	else
	{
		int bufPos;
		int64_t tempPos;
		vector<vector<float> >& initialBuffers = soundFileLoader->getInitialBuffers();
		float tempBuf[StreamingSound::NumBuffers][StreamingSound::BufferSize*2];

		for(i=0;i<StreamingSound::NumBuffers;++i)
		{
			bufPos = 0;
			while(bufPos < StreamingSound::BufferSize)
			{
				tempPos = (StreamingSound::BufferSize * i) + bufPos;

				bufPos = soundFileLoader->loadChunk(tempBuf[i],
													tempPos,
													(StreamingSound::BufferSize-bufPos) * 2,
													bufPos);
			}
		}

		//Convert mono files to stereo.
		if(soundFileLoader->getNumChannels() < 2)
		{
			for(i=0;i<StreamingSound::NumBuffers;++i)
			{
				for(j=(StreamingSound::BufferSize * 2)-2;j>=0;j-=2)
				{
					tempBuf[i][j] = tempBuf[i][j/2];
					tempBuf[i][j+1] = tempBuf[i][j];
				}
			}
		}

		//Just in case.
		for(i=0;i<(int)initialBuffers.size();++i)
			initialBuffers[i].clear();
		initialBuffers.clear();

		//Copy into soundFileLoader's initialBuffers.
		for(i=0;i<StreamingSound::NumBuffers;++i)
		{
			initialBuffers.push_back(vector<float>());

			for(j=0;j<(StreamingSound::BufferSize * 2);++j)
				initialBuffers[i].push_back(tempBuf[i][j]);
		}
	}
}

//------------------------------------------------------------------------------
bool SnakeballSounder::run()
{
	int i;

	while(!timeToStop)
	{
		for(i=0;i<NumSounds;++i)
		{
			if(sounds[i].getActive())
				sounds[i].threadUpdate();
		}

		//Wait a bit before looping.
		Thread::wait(62);
	}

	return true;
}

//------------------------------------------------------------------------------
void SnakeballSounder::playMusic(const std::wstring& sound)
{
	int i;

	//If it's a short sound, set it up accordingly.
	if(preLoadedBuffers.find(sound) != preLoadedBuffers.end())
	{
		if(music > -1)
			sounds[music].stopPlaying();

		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				music = i;
				sounds[i].setFade(true);
				sounds[i].setLoop(true);
				sounds[i].startPlaying(preLoadedBuffers[sound].buffer,
									   preLoadedBuffers[sound].size,
									   sound);

				break;
			}
		}
	}
	//Ditto if it's a longer file.
	else if(soundFiles.find(sound) != soundFiles.end())
	{
		if(music > -1)
			sounds[music].stopPlaying();

		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				music = i;
				sounds[i].setFade(true);
				sounds[i].setLoop(true);
				sounds[i].startPlaying(soundFiles[sound]);

				break;
			}
		}
	}
	else
		DebugStatement(L"SnakeballSounder: Could not find sound " << sound << " to play as music.");
}

//------------------------------------------------------------------------------
void SnakeballSounder::stopMusic()
{
	if(music > -1)
	{
		sounds[music].stopPlaying();
		music = -1;
	}
}

//------------------------------------------------------------------------------
void SnakeballSounder::triggerSound(const std::wstring& sound, float level, float pan)
{
	int i;

	//If it's a short sound, set it up accordingly.
	if(preLoadedBuffers.find(sound) != preLoadedBuffers.end())
	{
		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				sounds[i].setFade(false);
				sounds[i].setLoop(false);
				sounds[i].setLevel(level, false);
				sounds[i].setPan(pan);
				sounds[i].startPlaying(preLoadedBuffers[sound].buffer,
									   preLoadedBuffers[sound].size,
									   sound);

				break;
			}
		}
	}
	//Ditto if it's a longer file.
	else if(soundFiles.find(sound) != soundFiles.end())
	{
		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				sounds[i].setFade(false);
				sounds[i].setLoop(false);
				sounds[i].setLevel(level, false);
				sounds[i].setPan(pan);
				sounds[i].startPlaying(soundFiles[sound]);

				break;
			}
		}
	}
	else
		DebugStatement(L"SnakeballSounder: Could not find sound " << sound << " to play as a one-shot.");
}

//------------------------------------------------------------------------------
void SnakeballSounder::triggerEnvelope(float level, float pan)
{
	int i;

	for(i=0;i<NumSounds;++i)
	{
		if(envelopes[i].inc == 0.0f)
		{
			envelopes[i].inc = 0.02f;
			envelopes[i].index = 0.0f;
			envelopes[i].level = level;
			if(pan < 0.5f)
			{
				envelopes[i].left = 1.0f;
				envelopes[i].right = pan * 2.0f;
			}
			else
			{
				envelopes[i].left = 1.0f - ((pan-0.5f) * 2.0f);
				envelopes[i].right = 1.0f;
			}
			break;
		}
	}
}

//------------------------------------------------------------------------------
int SnakeballSounder::triggerLoop(const std::wstring& sound, float pan)
{
	int i;
	int retval = -1;

	//If it's a short sound, set it up accordingly.
	if(preLoadedBuffers.find(sound) != preLoadedBuffers.end())
	{
		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				sounds[i].setFade(false);
				sounds[i].setLoop(true);
				sounds[i].setPan(pan);
				sounds[i].startPlaying(preLoadedBuffers[sound].buffer,
									   preLoadedBuffers[sound].size,
									   sound);
				retval = i;

				break;
			}
		}
	}
	//Ditto if it's a longer file.
	else if(soundFiles.find(sound) != soundFiles.end())
	{
		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
			{
				sounds[i].setFade(false);
				sounds[i].setLoop(true);
				sounds[i].setPan(pan);
				sounds[i].startPlaying(soundFiles[sound]);
				retval = i;

				break;
			}
		}
	}
	else
		DebugStatement(L"SnakeballSounder: Could not find sound " << sound << " to play as a loop.");

	return retval;
}

//------------------------------------------------------------------------------
void SnakeballSounder::stopLoop(int index)
{
	if(sounds[index].getActive())
		sounds[index].stopPlaying();
}

//------------------------------------------------------------------------------
void SnakeballSounder::setNumChannels(int val)
{
	numChannels = val;
}

//------------------------------------------------------------------------------
void SnakeballSounder::setRedInputChannel(int val)
{
	redInputChannel = val;
}

//------------------------------------------------------------------------------
void SnakeballSounder::setBlueInputChannel(int val)
{
	blueInputChannel = val;
}

//------------------------------------------------------------------------------
void SnakeballSounder::setShouting(bool val)
{
	shouting = val;
}
